/*--------------------------------------------------------------------------+
$Id: ElementEditPartBase.java 8344 2013-08-05 08:11:09Z hoelzl $
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editpart;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getNodeBounds;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.draw2d.Figure;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.TreeSearch;
import org.eclipse.draw2d.XYLayout;
import org.eclipse.draw2d.geometry.Insets;
import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.layout.DefaultLayoutConstants;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.ui.editpart.figure.PrettyRoundedRectangle;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

/**
 * The base class for a (partial) diagram edit part, which is intended to be part of another
 * diagram.
 * 
 * @param <T>
 *            the type of the model element wrapped.
 * 
 * @author mou
 * @author $Author: hoelzl $
 * @version $Rev: 8344 $
 * @ConQAT.Rating GREEN Hash: E0E087D40CAF93B5E2B46E7B0CFE5E9D
 */
public abstract class SubDiagramEditPartBase<T extends ILayoutedModelElement & INamedCommentedElement>
		extends ElementEditPartBase<T> {

	/** The label layer for internal labels. */
	private Figure labelFigureLayer;

	/** Constructor. */
	protected SubDiagramEditPartBase(T modelObject) {
		super(modelObject);
	}

	/**
	 * <p>
	 * {@inheritDoc}
	 * </p>
	 * <p>
	 * Returns the connectors as children.
	 * </p>
	 */
	@Override
	protected List<?> getModelChildren() {
		List<Object> result = new ArrayList<Object>(modelElementHandler.getSubnodes(getModel()));
		result.addAll(modelElementHandler.getConnectors(getModel()));
		return result;
	}

	/** {@inheritDoc} */
	@Override
	protected IFigure createBaseFigure() {
		return new PrettyRoundedRectangle(0, new Insets(
				DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE / 2 - 2)) {
			/** {@inheritDoc} */
			@Override
			protected boolean useLocalCoordinates() {
				return true;
			}
		};
	}

	/** {@inheritDoc} */
	@Override
	protected Rectangle determineBaseFigureBounds() {
		Rectangle result = new Rectangle();
		for(Object o : getModelChildren()) {
			if(o instanceof ILayoutedModelElement) {
				Rectangle p = getLayoutedElementBounds((ILayoutedModelElement)o);
				result.union(p);
			}
		}
		result.resize(DEFAULT_GRID_SIZE, DEFAULT_GRID_SIZE);
		return result;
	}

	/**
	 * Determines the boundaries for a given {@link ILayoutedModelElement} in the context of this
	 * EditPart.
	 * 
	 * @param layouted
	 *            The {@link ILayoutedModelElement} to determine the boundaries for.
	 * @return {@link Rectangle} describing the boundaries.
	 */
	protected Rectangle getLayoutedElementBounds(ILayoutedModelElement layouted) {
		return getNodeBounds(layouted);
	}

	/** {@inheritDoc} */
	@Override
	protected Rectangle determineLabelFigureBounds() {
		Rectangle textBounds = labelFigure.getTextBounds();
		textBounds.y = DEFAULT_SHAPE_INSETS / 2;
		textBounds.x = DEFAULT_SHAPE_INSETS;
		textBounds.width = textBounds.width + 4 * DEFAULT_SHAPE_INSETS;
		textBounds.height = textBounds.height * 2;
		return textBounds;
	}

	/** {@inheritDoc} */
	@Override
	public Label createLabelFigure() {
		Label label = super.createLabelFigure();
		label.setLabelAlignment(PositionConstants.LEFT);
		return label;
	}

	/** {@inheritDoc} */
	@Override
	protected void registerLabelFigure(IFigure figure) {
		if(labelFigureLayer == null) {
			labelFigureLayer = new Figure() {
				/** {@inheritDoc} */
				@Override
				protected boolean useLocalCoordinates() {
					return true;
				}

				/** {@inheritDoc} */
				@Override
				public IFigure findFigureAt(int x, int y, TreeSearch search) {
					// transparent
					return null;
				}

				/** {@inheritDoc} */
				@Override
				public boolean isOpaque() {
					// transparent
					return false;
				}
			};
			labelFigureLayer.setLayoutManager(new XYLayout());
			super.registerLabelFigure(labelFigureLayer);
		}
		labelFigureLayer.add(figure);
	}

	/** {@inheritDoc} */
	@Override
	protected void refreshVisuals() {
		Rectangle r = determineBaseFigureBounds();
		labelFigureLayer.setBounds(r);
		labelFigureLayer.getParent().setConstraint(labelFigureLayer, r);

		super.refreshVisuals();
	}
}
