package org.fortiss.tooling.kernel.ui.internal;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getBundle;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getConfigurationElements;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.loadClass;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.warning;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.jface.resource.ImageDescriptor;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.Constraint;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsPackage;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.ToolingKernelUIActivator;
import org.fortiss.tooling.kernel.ui.extension.IConstraintVerifierUI;
import org.fortiss.tooling.kernel.ui.internal.introspection.items.ConstraintVerificationUIServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.ui.service.IConstraintVerificationUIService;
import org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;
import org.osgi.framework.Bundle;

/**
 * Implementation of {@link IConstraintVerificationService}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 5FB1D1BA3A20C0560809BA04FE487D4E
 */
public final class ConstraintVerificationUIService implements IConstraintVerificationUIService,
		IIntrospectiveKernelService {

	/** The singleton service instance. */
	private static ConstraintVerificationUIService INSTANCE = new ConstraintVerificationUIService();

	/** Stores the class to handler map. */
	protected final Map<String, IConstraintVerifierUI> handlerMap =
			new HashMap<String, IConstraintVerifierUI>();

	/** Returns the service instance. */
	public static ConstraintVerificationUIService getInstance() {
		return INSTANCE;
	}

	/** Activates logging to observe when constraints get outdated. */
	private static final boolean DEBUG = false;

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
		List<ITopLevelElement> tops = IPersistencyService.getInstance().getTopLevelElements();
		for(ITopLevelElement top : tops) {
			TreeIterator<EObject> tree = top.getRootModelElement().eAllContents();
			while(tree.hasNext()) {
				EObject next = tree.next();
				if(next instanceof Constraint) {
					install((Constraint)next);
				}
			}
		}
	}

	/** Registers the verifier UI part with the service. */
	@Override
	public void registerConstraintVerifierUI(IConstraintVerifierUI verifier) {
		addHandler(verifier.getID(), verifier);
	}

	/** Initializes the service by setting up the handler map. */
	public void initializeService() {
		setupHandlerMap();
	}

	/** Initializes the handler map from plugin extensions. */
	private void setupHandlerMap() {
		for(IConfigurationElement ce : getConfigurationElements(getExtensionPointName(),
				getConfigurationElementName())) {
			Bundle bundle = getBundle(ce);
			try {
				Class<?> handlerClass =
						loadClass(ce.getAttribute(getHandlerClassAttribute()), bundle);
				IConstraintVerifierUI handler =
						(IConstraintVerifierUI)handlerClass.getConstructor().newInstance();
				addHandler(handler.getID(), handler);
			} catch(Exception ex) {
				error(ToolingKernelActivator.getDefault(), ex.getMessage(), ex);
			}
		}
	}

	/** Adds the given handler to the map and the set. */
	protected void addHandler(String id, IConstraintVerifierUI handler) {
		IConstraintVerifierUI existingVerifier = handlerMap.get(id);
		if(existingVerifier != null) {
			warning(ToolingKernelActivator.getDefault(),
					"Encountered more than one handler registered with " + getExtensionPointName() +
							"(" + handler.getClass() + "," + existingVerifier.getClass() + ")");
		}
		handlerMap.put(id, handler);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service provides the UI complement of the \"pure\" constraint verification service." +
				"\nIt defines what should happen when it is requested that a status be opened," +
				"\nwhether a status can be opened, and it provides the facility to install on a constraint" +
				"\nthe various notifiers which are required for \"outdating\" detection." +
				"\n\nThis service is \"fed\" by constraint verifier UI's." +
				"\nConstraint verifiers define what should happen when it is requested that a status be opened," +
				"\nwhether a status can be opened, and what is the short description of the constraint status," +
				"\ne.g., in a context menu." +
				"\n\nThe class org.fortiss.tooling.kernel.ui.extension.base.ConstraintVerifierUIBase<T> should" +
				"\ngenerally provide a sufficient *initial* implementation. You are however strongly encouraged" +
				"\nto extend this class to provide a better feedback to the user (e.g., with detailed errors)." +
				"\n\nThe service extension point is '" + getExtensionPointName() + "'.";
	}

	/**
	 * @param constraint
	 * @return the first registered verifier which provides a GUI for the verifier of
	 *         <code>constraint</code>
	 */
	private IConstraintVerifierUI getFirstVerifier(Constraint constraint) {
		return getFirstVerifier(constraint.getConstraintTypeID());
	}

	/**
	 * @param id
	 * @return the first registered verifier which can verify a constraint of type <code>id</code>.
	 */
	protected IConstraintVerifierUI getFirstVerifier(String id) {
		IConstraintVerifierUI existingVerifier = handlerMap.get(id);
		if(existingVerifier == null) {
			LoggingUtils.error(ToolingKernelActivator.getDefault(),
					String.format("Cannot find verifier UI for constraint %s", id));
			return null;
		}
		return existingVerifier;
	}

	/** {@inheritDoc} */
	@Override
	public void openStatus(Constraint c) {
		IConstraintVerifierUI verifier = getFirstVerifier(c);
		verifier.openStatus(c, getStatus(c));
		ConstraintsUIUtils.triggerMarkersRefresh(c);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canOpen(Constraint c) {
		IConstraintVerifierUI verifier = getFirstVerifier(c);
		return verifier.canOpen(c, getStatus(c));
	}

	/** {@inheritDoc} */
	@Override
	public void install(Constraint constraint) {
		if(constraint.getVerificationStatus() != null &&
				constraint.getVerificationStatus().getConstraint() == null) {
			ITopLevelElement top =
					IPersistencyService.getInstance().getTopLevelElementFor(constraint);
			top.runAsCommand(() -> constraint.getVerificationStatus().setConstraint(constraint));
		}
		ConstraintsUIUtils.triggerMarkersRefresh(constraint);
		new ConstrainedAdapters(constraint);
	}

	/** Adapter for changes on constrained elements. */
	private class ConstrainedAdapters {

		/** Map containing the adapter of each constrained element. */
		Map<IConstrained, Adapter> constrainedAdapterMap;

		/** Constructor. */
		public ConstrainedAdapters(Constraint constraint) {
			constrainedAdapterMap = new HashMap<IConstrained, Adapter>();
			for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
				IConstrained constrained = cwc.getConstrained();
				Adapter adapter = new EContentAdapter() {
					@Override
					public void notifyChanged(Notification notification) {
						super.notifyChanged(notification);
						if(getStatus(constraint) == null || isTouch(notification)) {
							return;
						}
						if(constraintRemoved(notification, constrained)) {
							return;
						}
						constraintOutdated(constraint, notification);
					}
				};
				constrained.eAdapters().add(adapter);
				constrainedAdapterMap.put(constrained, adapter);
			}
		}

		/**
		 * @param constraint
		 * @param notification
		 *            Checks if <code>notification</code> denotes a change entailing that
		 *            <code>constraint</code> is outdated. If so updates the status to outdated or
		 *            triggers an automatic check if the <code>autocheck</code> field of the
		 *            constraint is <code>true</code>.
		 * @return <code>true</code> iff <code>notification</code> indeed entailed that
		 *         <code>constraint</code> is outdated.
		 */
		private boolean constraintOutdated(Constraint constraint, Notification notification) {
			if(skipStatusChangeNotification(constraint, notification)) {
				return false;
			}
			if(!IConstraintVerificationService.getInstance().isUpToDate(constraint)) {
				if(DEBUG) {
					String msg = "Constraint " + constraint.toString() + " is outdated.";
					LoggingUtils.info(ToolingKernelUIActivator.getDefault(), msg);
				}
				getFirstVerifier(constraint).onOutdate(constraint);
				return true;
			}
			return false;
		}

		/**
		 * @param constraint
		 * @param notification
		 * @return true iff <code>notification</code> denotes a change of the
		 *         <code>constraint</code>'s status.
		 */
		private boolean skipStatusChangeNotification(Constraint constraint,
				Notification notification) {
			String n = ConstraintsPackage.eINSTANCE.getConstraint_VerificationStatus().getName();
			if(notification.getNotifier() != null && notification.getNotifier().equals(constraint)) {
				if(notification.getFeature() instanceof EReference) {
					return(n.equals(((ENamedElement)notification.getFeature()).getName()));
				}
			}
			return false;
		}

		/**
		 * @param notification
		 * @param constrained
		 * @return true if the notification is a consequence of a constraint of
		 *         <code>constrained</code> having been removed.
		 */
		private boolean constraintRemoved(Notification notification, IConstrained constrained) {
			if(notification.getEventType() != Notification.REMOVE) {
				return false;
			}
			if(!notification.getNotifier().equals(constrained)) {
				return false;
			}
			if(notification.getFeature() instanceof EReference) {
				String name = ((ENamedElement)notification.getFeature()).getName();
				if(name.equals(ConstraintsPackage.eINSTANCE.getIConstrained_Constraints().getName())) {
					constrained.eAdapters().remove(constrainedAdapterMap.get(constrained));
					constrainedAdapterMap.remove(constrained);
					return true;
				}
			}
			return true;
		}
	}

	/**
	 * @param notification
	 * @return true if <code>notification</code> is not relevant and should be skipped
	 */
	private static boolean isTouch(Notification notification) {
		// Notifications which do not notify about a model change.
		if(notification.isTouch()) {
			return true;
		}
		// Notifications initiated by the the AF3 kernel which are only GUI related.
		if(notification.getEventType() == EcoreUtils.KERNEL_EMF_EVENT_TYPE_COUNT) {
			return true;
		}
		return false;
	}

	/** Returns the extension point name. */
	protected String getExtensionPointName() {
		return "org.fortiss.tooling.kernel.ui.constraintVerifierUI";
	}

	/** Returns the configuration element name. */
	protected String getConfigurationElementName() {
		return "constraintVerifierUI";
	}

	/** Returns the handler class attribute name. */
	protected String getHandlerClassAttribute() {
		return "constraintVerifierUI";
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ConstraintVerificationUIServiceIntrospectionDetailsItem(handlerMap);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Constraint Verification UI Service";
	}

	/** Activated constraints per top loevel element. */
	private static Map<ITopLevelElement, Set<String>> activatedConstraints = new HashMap<>();

	/** {@inheritDoc} */
	@Override
	public IConstraintVerificationStatus getStatus(Constraint constraint) {
		ITopLevelElement top = IPersistencyService.getInstance().getTopLevelElementFor(constraint);
		boolean isActivated = false;
		Set<String> set = activatedConstraints.get(top);
		if(set != null) {
			for(String id : set) {
				if(id.equals(constraint.getConstraintTypeID())) {
					isActivated = true;
				}
			}
		}
		return isActivated ? constraint.getVerificationStatus() : null;
	}

	/** {@inheritDoc} */
	@Override
	public Set<String> getAvailableConstraintTypes() {
		return handlerMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public void activate(String id, IConstraintContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		if(!activatedConstraints.containsKey(top)) {
			activatedConstraints.put(top, new HashSet<>());
		}
		activatedConstraints.get(top).add(id);
		IConstraintVerifierUI verifier = getFirstVerifier(id);
		verifier.onActivate(cstrContainer);
		for(Constraint c : cstrContainer.getConstraints()) {
			if(id.equals(c.getConstraintTypeID())) {
				if(!IConstraintVerificationService.getInstance().isUpToDate(c) ||
						c.getVerificationStatus() == null ||
						c.getVerificationStatus() instanceof OutdatedVerificationStatus) {
					verifier.onOutdate(c);
				}
			}
		}
		ConstraintsUIUtils.triggerMarkerRefresh(id, cstrContainer);
	}

	/** {@inheritDoc} */
	@Override
	public void deactivate(String id, IConstraintContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		activatedConstraints.get(top).remove(id);
		ConstraintsUIUtils.triggerMarkerRefresh(id, cstrContainer);
	}

	/** {@inheritDoc} */
	@Override
	public String getDescription(String id) {
		IConstraintVerifierUI verifier = getFirstVerifier(id);
		return verifier == null ? null : verifier.getDescription();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public ImageDescriptor getIconImageDescriptor(String id) {
		IConstraintVerifierUI verifier = getFirstVerifier(id);
		return verifier == null ? null : verifier.getIconImageDescriptor();
	}

	/** {@inheritDoc} */
	@Override
	public boolean shallDisplayAsWarning(Constraint constraint) {
		return getFirstVerifier(constraint).displayAsWarning();
	}
}
