package org.fortiss.tooling.kernel.service;

import java.util.List;
import java.util.function.Consumer;

import org.fortiss.tooling.kernel.extension.IConstraintVerifier;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.internal.ConstraintVerificationService;
import org.fortiss.tooling.kernel.model.constraints.Constraint;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;

/**
 * The verification service allows to register "constraint verifiers", e.g.,
 * a test suite runner, a temporal logic verifier, a WCET checker, a requirement review system.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 4B4CDB48E5D52B2656D7453E8E87A860
 */
public interface IConstraintVerificationService {

	/** Returns the service instance. */
	public static IConstraintVerificationService getInstance() {
		return ConstraintVerificationService.getInstance();
	}

	/**
	 * @param constraint
	 *            Verifies <code>constraint</code>. The framework always runs the verification in a
	 *            command. Should not yield any "useless" change because verification should always
	 *            update the constraint status and the checksums anyways.
	 */
	void verify(Constraint constraint);

	/**
	 * @param constraint
	 *            Cancels the last verification of <code>constraint</code>.
	 */
	void cancel(Constraint constraint);

	/**
	 * @param constraint
	 * @return <code>true</code> if <code>constraint</code> is up to date.
	 *         This does not just look at the status but computes a diff between the constrained
	 *         element and the one for which the constraint was last time checked. In the standard
	 *         case, it is enough to call IConstraintVerificationUIService.getStatus and see if the
	 *         result is an instance of {@link OutdatedVerificationStatus}.You should use
	 *         {isUpToDate} *only if IConstraintVerificationUIService.getStatus does not provide the
	 *         information you want*.
	 */
	boolean isUpToDate(Constraint constraint);

	/**
	 * @param constraint
	 * @return A list of possible automatic fixes for <code>constraint</code>. A constraint should
	 *         be candidate to fixing if it is anything but successful or outdated.
	 */
	List<IFix> fixes(Constraint constraint);

	/** Interface for a "fix". */
	public interface IFix {
		/**
		 * Description of the fix. Useful to allow the user distinguish the fix in case there are
		 * several.
		 */
		String getDescription();

		/**
		 * @param status
		 *            Fixes (the constraint of) <code>status</code>.
		 */
		void runFix(IConstraintVerificationStatus status);
	}

	/** Utils. We store them to avoid having again thousands of different classes for one topic. */

	/** Base for fixes. */
	public static class FixBase implements IFix {
		/** The description. */
		private String description;

		/** The fix itself. */
		private Consumer<IConstraintVerificationStatus> fix;

		/** Constructor. */
		public FixBase(String description, Consumer<IConstraintVerificationStatus> fix) {
			this.description = description;
			this.fix = fix;
		}

		/** {@inheritDoc} */
		@Override
		public String getDescription() {
			return description;
		}

		/** {@inheritDoc} */
		@Override
		public void runFix(IConstraintVerificationStatus status) {
			fix.accept(status);
		}
	}

	/**
	 * @param constraint
	 * @param constrained
	 * @param index
	 *            Defines <code>constrained</code> as the <code>index</code>nth constrained model
	 *            element of <code>constraint</code>.
	 *            Leaves the other elements as they are if they exist, fill them with
	 *            <code>null</code> till <code>index</code> otherwise.
	 */
	void setConstrained(Constraint constraint, IConstrained constrained, int index);

	/**
	 * @param top
	 * @param id
	 *            Completes <code>top</code> with missing constraints of type <code>id</code>, if
	 *            applicable.
	 * @return List of added constraints.
	 * 
	 */
	List<Constraint> addMissingConstraints(ITopLevelElement top, String id);

	/** Registers the given verifier with the service. */
	void registerConstraintVerifier(IConstraintVerifier verifier);
}
